import beniget
import gast as ast
import os
import sys
import warnings

from importlib.util import resolve_name
from collections import defaultdict
from memestra.caching import Cache, CacheKeyFactory, RecursiveCacheKeyFactory
from memestra.caching import Format
from memestra.utils import resolve_module

_defs = ast.AsyncFunctionDef, ast.ClassDef, ast.FunctionDef

class DeprecatedStar(object):
    'Representation of a deprecated node imported through *'

    def __init__(self, name, dnode):
        assert dnode.name == '*'
        self.name = name
        self.dnode = dnode

def symbol_name(sym):
    if getattr(sym, 'asname', None):
        return sym.asname
    if hasattr(sym, 'name'):
        return sym.name
    if hasattr(sym, 'id'):
        return sym.id
    raise NotImplementedError(sym)

def make_deprecated(node, reason=None):
    return (node, reason)

def store_deprecated(name, reason):
    return name if reason is None else '{}:{}'.format(name, reason)

def load_deprecated(entry):
    if ':' in entry:
        return entry.split(':', maxsplit=1)
    else:
        return entry, None

class SilentDefUseChains(beniget.DefUseChains):

    def unbound_identifier(self, name, node):
        pass


class SilentDefUseChains(beniget.DefUseChains):

    def unbound_identifier(self, name, node):
        pass


class ImportResolver(ast.NodeVisitor):

    def __init__(self, decorator, reason_keyword, search_paths=None,
                 recursive=False, parent=None, pkg_name=None,
                 cache_dir=None):
        '''
        Create an ImportResolver that finds deprecated identifiers.

        A deprecated identifier is an identifier which is decorated
        by `decorator', or which uses a deprecated identifier.

        if `recursive' is greater than 0, it considers identifiers
        from imported module, with that depth in the import tree.

        `parent' is used internally to handle imports.
        '''
        self.deprecated = None
        self.decorator = tuple(decorator)
        self.search_paths = search_paths
        self.recursive = recursive
        self.reason_keyword = reason_keyword
        self.pkg_name = pkg_name
        if parent:
            self.cache = parent.cache
            self.visited = parent.visited
            self.key_factory = parent.key_factory
        else:
            self.cache = Cache(cache_dir=cache_dir)
            self.visited = set()
            if recursive:
                self.key_factory = RecursiveCacheKeyFactory()
            else:
                self.key_factory = CacheKeyFactory()

    def load_deprecated_from_module(self, module_name, level=None):
        # level may be none when it's taken from the ImportFrom node
        if level is None:
            level = 0

        # update module/pkg based on level
        rmodule_name = '.' * level + module_name

        # perform the module lookup
        try:
            module_name = resolve_name(rmodule_name, self.pkg_name)
        except (ImportError, ValueError):
            return None
        module_path = resolve_module(module_name, self.search_paths)

        # hopefully a module was found
        if module_path is None:
            return None

        module_key = self.key_factory(module_path)

        # either find it in the cache
        if module_key in self.cache:
            data = self.cache[module_key]
            if data['version'] == Format.version:
                return dict(load_deprecated(entry)
                            for entry in data['deprecated'])
            elif data['generator'] == 'manual':
                warnings.warn(
                    ("skipping module {} because it has an obsolete, "
                     "manually generated, cache file: {}")
                    .format(module_name,
                            module_key.module_hash))
                return {}

        # or fill a new entry

        # To avoid loop, mark the key as in process
        self.cache[module_key] = {'generator': 'manual',
                                  'deprecated': []}

        with open(module_path) as fd:
            try:
                module = ast.parse(fd.read())
            except UnicodeDecodeError:
                return []
            duc = SilentDefUseChains()
            duc.visit(module)
            anc = beniget.Ancestors()
            anc.visit(module)

            # Collect deprecated functions
            if self.recursive and module_path not in self.visited:
                self.visited.add(module_path)
                current_pkg = ".".join(module_name.split('.')[:-1])
                resolver = ImportResolver(self.decorator,
                                          self.reason_keyword,
                                          self.search_paths,
                                          self.recursive,
                                          parent=self,
                                          pkg_name=current_pkg)
                resolver.visit(module)
                deprecated_imports = [make_deprecated(d, reason)
                                      for _, _, d, reason in
                                      resolver.get_deprecated_users(duc, anc)]
            else:
                deprecated_imports = []
            deprecated = self.collect_deprecated(module, duc, anc,
                                                 pkg_name=module_name)
            deprecated.update(deprecated_imports)
            dl = {symbol_name(d[0]): d[1] for d in deprecated if d is not None}
            data = {'generator': 'memestra',
                    'deprecated': [store_deprecated(d, dl[d]) for d in
                                   sorted(dl)]}
            self.cache[module_key] = data
            return dl

    def get_deprecated_users(self, defuse, ancestors):
        deprecated_uses = []
        visited = set()
        worklist = list(self.deprecated)
        while worklist:
            deprecated_node, reason = worklist.pop()
            if deprecated_node in visited:
                continue
            visited.add(deprecated_node)

            # special node: an imported name
            if isinstance(deprecated_node, ast.alias):
                deprecated_uses.append((deprecated_node,
                                        ancestors.parent(deprecated_node),
                                        deprecated_node, reason))

            else:
                # special node: a node imported from *; in that case we do
                # selective matching
                if isinstance(deprecated_node, DeprecatedStar):
                    def filter_out(user):
                        return user.name() != deprecated_node.name
                    deprecated = deprecated_node.dnode
                # regular symbol processing
                else:
                    def filter_out(user):
                        return False
                    deprecated = deprecated_node

                for user in defuse.chains[deprecated].users():
                    if filter_out(user):
                        continue
                    user_ancestors = [n
                                      for n in ancestors.parents(user.node)
                                      if isinstance(n, _defs)]
                    if any(f in self.deprecated for f in user_ancestors):
                        continue

                    user_ancestor = user_ancestors[-1] if user_ancestors else user.node
                    deprecated_uses.append((deprecated, user.node,
                                            user_ancestor,
                                            reason))
                    if self.recursive and isinstance(user_ancestor, _defs):
                        worklist.append((user_ancestor, reason))
        return deprecated_uses

    def visit_Import(self, node):
        for alias in node.names:
            deprecated = self.load_deprecated_from_module(alias.name)
            if deprecated is None:
                continue

            for user in self.def_use_chains.chains[alias].users():
                parent = self.ancestors.parents(user.node)[-1]
                if isinstance(parent, ast.Attribute):
                    if parent.attr in deprecated:
                        reason = deprecated[parent.attr]
                        self.deprecated.add(make_deprecated(parent, reason))

    def visit_ImportFrom(self, node):
        if node.module is None:
            return None
        deprecated = self.load_deprecated_from_module(node.module,
                                                      level=node.level)
        if deprecated is None:
            return

        aliases = [alias.name for alias in node.names]

        try:
            # if we're importing *, pick all deprecated items
            star_alias_index = aliases.index('*')
            star_alias = node.names[star_alias_index]

            def alias_extractor(deprec):
                return star_alias

        except ValueError:
            # otherwise only pick the imported one

            def alias_extractor(deprec):
                try:
                    index = aliases.index(deprec)
                    alias = node.names[index]
                    return alias
                except ValueError:
                    return None

        for deprec, reason in deprecated.items():
            deprec_alias = alias_extractor(deprec)
            if deprec_alias is None:
                continue

            for user in self.def_use_chains.chains[deprec_alias].users():
                self.deprecated.add(make_deprecated(user.node, reason))

    def visit_Module(self, node):
        duc = SilentDefUseChains()
        duc.visit(node)
        self.def_use_chains = duc

        ancestors = beniget.Ancestors()
        ancestors.visit(node)
        self.ancestors = ancestors

        self.deprecated = self.collect_deprecated(node, duc, ancestors)
        self.generic_visit(node)

    def collect_deprecated(self, node, duc, ancestors, pkg_name=None):
        deprecated = set()

        for dlocal in duc.locals[node]:
            dnode = dlocal.node
            if not isinstance(dnode, ast.alias):
                continue

            original_path = tuple(dnode.name.split('.'))

            # for imports with a "from" clause, such as
            #
            #   from foo import bar
            #
            # the AST alias will be just `bar`, but we want any functions
            # defined as such:
            #
            # @bar
            # def foo(): pass
            #
            # to be picked up when `foo.bar` is used as the target decorator. we
            # check if the parent of the alias is an ImportFrom node and fix the
            # original path to be fully qualified here. In the example above, it
            # becomes `foo.bar` instead of just `bar`.
            alias_parent = ancestors.parents(dnode)[-1]
            if isinstance(alias_parent, ast.ImportFrom):
                module = "."
                # A module can be None if a relative import from "." occurs
                if alias_parent.module is not None:
                    module = resolve_name(alias_parent.module, self.pkg_name)

                original_path = tuple(module.split('.')) + original_path

            nbterms = len(original_path)

            if original_path == self.decorator[:nbterms]:
                for user in dlocal.users():
                    parents = list(ancestors.parents(user.node))
                    attrs = list(reversed(self.decorator[nbterms:]))
                    while attrs and parents:
                        attr = attrs[-1]
                        parent = parents.pop()
                        if not isinstance(parent, (ast.Attribute)):
                            break
                        if parent.attr != attr:
                            break
                        attrs.pop()

                    # path parsing fails if some attr left
                    if attrs:
                        continue

                    # Only handle decorators attached to a def
                    self.extract_decorator_from_parents(
                        parents,
                        deprecated)


        if not self.recursive:
            return deprecated

        # In recursive mode, we consider deprecated any imported
        # deprecated symbol.
        for dlocal in duc.locals[node]:
            dnode = dlocal.node
            if not isinstance(dnode, ast.alias):
                continue
            alias_parent = ancestors.parents(dnode)[-1]
            if isinstance(alias_parent, ast.ImportFrom):
                if not alias_parent.module:
                    continue
                resolver = ImportResolver(self.decorator,
                                          self.reason_keyword,
                                          self.search_paths,
                                          self.recursive,
                                          self,
                                          pkg_name)

                imported_deprecated = resolver.load_deprecated_from_module(
                    alias_parent.module,
                    level=alias_parent.level)
                if not imported_deprecated:
                    continue
                if dnode.name in imported_deprecated:
                    dinfo = make_deprecated(dnode,
                                            imported_deprecated[dnode.name])
                    deprecated.add(dinfo)
                elif dnode.name == '*':
                    for name, reason in imported_deprecated.items():
                        dinfo = make_deprecated(DeprecatedStar(name, dnode),
                                                reason)
                        deprecated.add(dinfo)

        return deprecated

    def extract_decorator_from_parents(self, parents, deprecated):
        parent = parents[-1]
        if isinstance(parent, _defs):
            deprecated.add(make_deprecated(parent))
            return
        if len(parents) == 1:
            return
        parent_p = parents[-2]
        if isinstance(parent, ast.Call) and isinstance(parent_p, _defs):
            reason = None
            # Output only the specified reason with the --reason-keyword flag
            if not parent.keywords and parent.args:
                reason = parent.args[0].value
            for keyword in parent.keywords:
                if self.reason_keyword == keyword.arg:
                    reason = keyword.value.value
            deprecated.add(make_deprecated(parent_p, reason))
            return

def prettyname(node):
    if isinstance(node, _defs):
        return node.name
    if isinstance(node, ast.alias):
        return node.asname or node.name
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.alias):
        return node.asname or node.name
    if isinstance(node, ast.Attribute):
        return prettyname(node.value) + '.' + node.attr
    return repr(node)


def memestra(file_descriptor, decorator, reason_keyword,
             search_paths=None, recursive=False, cache_dir=None):
    '''
    Parse `file_descriptor` and returns a list of
    (function, filename, line, colno) tuples. Each elements
    represents a code location where a deprecated function is used.
    A deprecated function is a function flagged by `decorator`, where
    `decorator` is a tuple representing an import path,
    e.g. (module, attribute)

    If `recursive` is set to `True`, deprecated use are
    checked recursively throughout the *whole* module import tree. Otherwise,
    only one level of import is checked.
    '''

    assert not isinstance(decorator, str) and \
           len(decorator) > 1, "decorator is at least (module, attribute)"

    module = ast.parse(file_descriptor.read())
    # Collect deprecated functions
    resolver = ImportResolver(decorator, reason_keyword, search_paths,
                              recursive, cache_dir=cache_dir)
    resolver.visit(module)

    ancestors = resolver.ancestors
    duc = resolver.def_use_chains

    # Find their users
    formated_deprecated = []
    for deprecated_info in resolver.get_deprecated_users(duc, ancestors):
        deprecated_node, user_node, _, reason = deprecated_info
        formated_deprecated.append((prettyname(deprecated_node),
                               getattr(file_descriptor, 'name', '<>'),
                               user_node.lineno,
                               user_node.col_offset,
                               reason))
    formated_deprecated.sort()
    return formated_deprecated


def run():

    import argparse
    from pkg_resources import iter_entry_points

    parser = argparse.ArgumentParser(description='Check decorator usage.')
    parser.add_argument('--decorator', dest='decorator',
                        default='deprecated.deprecated',
                        help='Path to the decorator to check')
    parser.add_argument('input', type=argparse.FileType('r'),
                        help='file to scan')
    parser.add_argument('--reason-keyword', dest='reason_keyword',
                        default='reason',
                        action='store',
                        help='Specify keyword for deprecation reason')
    parser.add_argument('--cache-dir', dest='cache_dir',
                        default=None,
                        action='store',
                        help='The directory where the cache is located')
    parser.add_argument('--recursive', dest='recursive',
                        action='store_true',
                        help='Traverse the whole module hierarchy')

    dispatcher = defaultdict(lambda: memestra)
    for entry_point in iter_entry_points(group='memestra.plugins', name=None):
        entry_point.load()(dispatcher)

    args = parser.parse_args()

    _, extension = os.path.splitext(args.input.name)

    # Add the directory of the python file to the list of import paths
    # to search.
    search_paths = [os.path.dirname(os.path.abspath(args.input.name))]

    deprecate_uses = dispatcher[extension](args.input,
                                           args.decorator.split('.'),
                                           args.reason_keyword,
                                           search_paths,
                                           args.recursive,
                                           args.cache_dir)

    for fname, fd, lineno, colno, reason in deprecate_uses:
        formatted_reason = ""
        if reason:
            formatted_reason = " - {}".format(reason)
        print("{} used at {}:{}:{}{}".format(fname, fd, lineno, colno + 1, formatted_reason))


if __name__ == '__main__':
    run()



import gast as ast
import re


def docparse(module_path, pattern):

    with open(module_path) as fd:
        content = fd.read()
    tree = ast.parse(content)
    def_types = (ast.AsyncFunctionDef,
                 ast.FunctionDef,
                 ast.ClassDef)
    flags = re.DOTALL | re.MULTILINE

    deprecated = []
    for stmt in tree.body:
        if not isinstance(stmt, def_types):
            continue
        fst_stmt = stmt.body[0]
        if not isinstance(fst_stmt, ast.Expr):
            continue
        if not isinstance(fst_stmt.value, ast.Constant):
            continue
        cst = fst_stmt.value
        if not isinstance(cst.value, str):
            continue
        if re.match(pattern, cst.value, flags):
            deprecated.append(stmt.name)
    return deprecated



import os
import sys
from importlib import util
from importlib.abc import SourceLoader

def _resolve_module(module_name, additional_search_paths=None):
    if additional_search_paths is None:
        additional_search_paths = []

    if module_name is None or module_name == "__main__":
        return None, None

    parent, _, module_name = module_name.rpartition(".")
    search_path = sys.path + additional_search_paths

    if parent != "":
        parent_spec, search_path = _resolve_module(parent)
        if parent_spec is None:
            return None, None

    try:
        spec = None
        for finder in sys.meta_path:
            try:
                spec = finder.find_spec(module_name, search_path)
            except AttributeError:
                pass

            if spec is not None:
                break

        origin = spec.origin if spec else None
        if spec is None or spec.origin is None:
            return None, None
        if isinstance(spec.loader, SourceLoader):
            return spec.origin, spec.submodule_search_locations
        else:
            return None, None

    except ModuleNotFoundError:
        return None, []

def resolve_module(module_name, additional_search_paths=None):
    return _resolve_module(module_name, additional_search_paths)[0]



import re
from nbconvert.preprocessors import Preprocessor
from memestra import nbmemestra
from traitlets import Tuple


class MemestraDeprecationChecker(Preprocessor):
    """A memestra-based preprocessor that checks code for places where deprecated functions are called"""

    decorator = Tuple(help="path to import the deprecation decorator, e.g. ('module', 'attribute')").tag(config=True)

    def preprocess(self, nb, resources):
        self.log.info("Using decorator '%s.%s' to check deprecated functions", self.decorator[0], self.decorator[1])
        deprecations = {}
        for d in nbmemestra.nbmemestra_from_nbnode(nb, self.decorator):
            code_cell_i = int(re.search(r"\[(\d+)\]", d[1]).group(1))
            deprecation = deprecations.get(code_cell_i, [])
            deprecation.append(d)
            deprecations[code_cell_i] = deprecation
        # filter out non-code cell as it is done in nbmemestra
        code_cell_i = -1
        for cell in nb.cells:
            if cell['cell_type'] != 'code':
                continue
            code_cell_i += 1
            if code_cell_i not in deprecations:
                continue
            for d in deprecations[code_cell_i]:
                line, col = d[2:4]
                outputs = cell.get('outputs', [])
                message = ('On line {}:\n'
                           '{}\n'
                           '{}^\n'
                           'Warning: call to deprecated function {}'.format(
                               line,
                               cell['source'].split('\n')[line - 1],
                               ' ' * col,
                               d[0]
                          ))

                outputs.append({'output_type': 'stream', 'name': 'stderr', 'text': message})
                cell['outputs'] = outputs
        return nb, resources



import os
import hashlib
import yaml

# not using gast because we only rely on Import and ImportFrom, which are
# portable. Not using gast prevents an extra costly conversion step.
import ast

from memestra.docparse import docparse
from memestra.utils import resolve_module


class DependenciesResolver(ast.NodeVisitor):
    '''
    Traverse a module an collect statically imported modules
    '''


    def __init__(self):
        self.result = set()

    def add_module(self, module_name):
        module_path = resolve_module(module_name)
        if module_path is not None:
            self.result.add(module_path)

    def visit_Import(self, node):
        for alias in node.names:
            self.add_module(alias.name)

    def visit_ImportFrom(self, node):
        self.add_module(node.module)

    # All members below are specialized in order to improve performance:
    # It's useless to traverse leaf statements and expression when looking for
    # an import.

    def visit_stmt(self, node):
        pass

    visit_Assign = visit_AugAssign = visit_AnnAssign = visit_Expr = visit_stmt
    visit_Return = visit_Print = visit_Raise = visit_Assert = visit_stmt
    visit_Pass = visit_Break = visit_Continue = visit_Delete = visit_stmt
    visit_Global = visit_Nonlocal = visit_Exec = visit_stmt

    def visit_body(self, node):
        for stmt in node.body:
            self.visit(stmt)

    visit_FunctionDef = visit_ClassDef = visit_AsyncFunctionDef = visit_body
    visit_With = visit_AsyncWith = visit_body

    def visit_orelse(self, node):
        for stmt in node.body:
            self.visit(stmt)
        for stmt in node.orelse:
            self.visit(stmt)

    visit_For = visit_While = visit_If = visit_AsyncFor = visit_orelse

    def visit_Try(self, node):
        for stmt in node.body:
            self.visit(stmt)
        for stmt in node.orelse:
            self.visit(stmt)
        for stmt in node.finalbody:
            self.visit(stmt)


class Format(object):

    version = 1

    fields = (('version', lambda: Format.version),
              ('name', str),
              ('deprecated', list),
              ('generator', lambda: 'manual'))

    generators = {'memestra', 'manual'}

    @staticmethod
    def setdefaults(data, **defaults):
        for field, default in Format.fields:
            data.setdefault(field, defaults.get(field, default()))

    @staticmethod
    def check(data):
        Format.check_keys(data)
        for k, _ in Format.fields:
            getattr(Format, 'check_{}'.format(k))(data)

    @staticmethod
    def check_name(data):
        if not isinstance(data['name'], str):
            raise ValueError("name must be an str")

    @staticmethod
    def check_keys(data):
        data_keys = set(data.keys())
        format_keys = {k for k, _ in Format.fields}
        if not data_keys == format_keys:
            difference = data_keys.difference(format_keys)
            raise ValueError("Invalid field{}: {}".format(
                's' * (len(difference) > 1),
                ', '.join(sorted(difference))))

    @staticmethod
    def check_version(data):
        if data['version'] != Format.version:
            raise ValueError(
                "Invalid version, should be {}".format(Format.version))

    @staticmethod
    def check_generator(data):
        if data['generator'] not in Format.generators:
            raise ValueError(
                "Invalid generator, should be one of {}"
                .format(', '.join(sorted(Format.generators))))

    @staticmethod
    def check_deprecated(data):
        deprecated = data['deprecated']
        if not isinstance(deprecated, list):
            raise ValueError("deprecated must be a list")
        if not all(isinstance(of, str) for of in deprecated):
            raise ValueError("deprecated must be a list of string")


class CacheKeyFactoryBase(object):
    def __init__(self, keycls):
        self.keycls = keycls
        self.created = dict()

    def __call__(self, module_path):
        if module_path in self.created:
            return self.created[module_path]
        else:
            self.created[module_path] = None  # creation in process
            key = self.keycls(module_path, self)
            self.created[module_path] = key
            return key

    def get(self, *args):
        return self.created.get(*args)


class CacheKeyFactory(CacheKeyFactoryBase):
    '''
    Factory for non-recursive keys.
    Only the content of the module is taken into account
    '''

    class CacheKey(object):

        def __init__(self, module_path, _):
            self.name, _ = os.path.splitext(os.path.basename(module_path))
            with open(module_path, 'rb') as fd:
                module_content = fd.read()
                module_hash = hashlib.sha256(module_content).hexdigest()
                self.module_hash = module_hash

    def __init__(self):
        super(CacheKeyFactory, self).__init__(CacheKeyFactory.CacheKey)


class RecursiveCacheKeyFactory(CacheKeyFactoryBase):
    '''
    Factory for recursive keys.
    This take into account the module content, and the content of *all* imported
    module. That way, a change in the module hierarchy implies a change in the
    key.
    '''

    class CacheKey(object):

        def __init__(self, module_path, factory):
            assert module_path not in factory.created or factory.created[module_path] is None

            self.name, _ = os.path.splitext(os.path.basename(module_path))
            with open(module_path, 'rb') as fd:
                module_content = fd.read()

                code = ast.parse(module_content)
                dependencies_resolver = DependenciesResolver()
                dependencies_resolver.visit(code)

                new_deps = []
                for dep in dependencies_resolver.result:
                    if factory.get(dep, 1) is not None:
                        new_deps.append(dep)

                module_hash = hashlib.sha256(module_content).hexdigest()

                hashes = [module_hash]

                for new_dep in sorted(new_deps):
                    try:
                        new_dep_key = factory(new_dep)
                    # FIXME: this only happens on windows, maybe we could do
                    # better?
                    except UnicodeDecodeError:
                        continue
                    hashes.append(new_dep_key.module_hash)

                self.module_hash = hashlib.sha256("".join(hashes).encode("ascii")).hexdigest()

    def __init__(self):
        super(RecursiveCacheKeyFactory, self).__init__(RecursiveCacheKeyFactory.CacheKey)


class Cache(object):

    def __init__(self, cache_dir=None):
        if cache_dir is not None:
            self.cachedir = cache_dir
        else:
            xdg_config_home = os.environ.get('XDG_CONFIG_HOME', None)
            if xdg_config_home is None:
                user_config_dir = '~'
                memestra_dir = '.memestra'
            else:
                user_config_dir = xdg_config_home
                memestra_dir = 'memestra'
            self.cachedir = os.path.expanduser(os.path.join(user_config_dir,
                                                            memestra_dir))

        os.makedirs(self.cachedir, exist_ok=True)

    def _get_path(self, key):
        return os.path.join(self.cachedir, key.module_hash)

    def __contains__(self, key):
        cache_path = self._get_path(key)
        return os.path.isfile(cache_path)

    def __getitem__(self, key):
        cache_path = self._get_path(key)
        with open(cache_path, 'r') as yaml_fd:
            return yaml.load(yaml_fd, Loader=yaml.SafeLoader)

    def __setitem__(self, key, data):
        data = data.copy()
        Format.setdefaults(data, name=key.name)
        Format.check(data)
        cache_path = self._get_path(key)
        with open(cache_path, 'w') as yaml_fd:
            yaml.dump(data, yaml_fd)

    def keys(self):
        return os.listdir(self.cachedir)

    def items(self):
        for key in self.keys():
            cache_path = os.path.join(self.cachedir, key)
            with open(cache_path, 'r') as yaml_fd:
                yield key, yaml.load(yaml_fd, Loader=yaml.SafeLoader)

    def clear(self):
        count = 0
        for key in self.keys():
            cache_path = os.path.join(self.cachedir, key)
            os.remove(cache_path)
            count += 1
        return count


def run_set(args):
    data = {'generator': 'manual',
            'deprecated': args.deprecated}
    cache = Cache(cache_dir=args.cache_dir)
    if args.recursive:
        key_factory = RecursiveCacheKeyFactory()
    else:
        key_factory = CacheKeyFactory()
    key = key_factory(args.input)
    cache[key] = data


def run_list(args):
    cache = Cache(cache_dir=args.cache_dir)
    for k, v in cache.items():
        print('{}: {} ({})'.format(k, v['name'], len(v['deprecated'])))


def run_clear(args):
    cache = Cache(cache_dir=args.cache_dir)
    nb_cleared = cache.clear()
    print('Cache cleared, {} element{} removed.'.format(nb_cleared, 's' *
                                                        (nb_cleared > 1)))


def run_docparse(args):
    deprecated = docparse(args.input, args.pattern)

    cache = Cache(cache_dir=args.cache_dir)
    if args.recursive:
        key_factory = RecursiveCacheKeyFactory()
    else:
        key_factory = CacheKeyFactory()
    key = key_factory(args.input)
    data = {'deprecated': deprecated,
            'generator': 'manual'}
    cache[key] = data
    if args.verbose:
        print("Found {} deprecated identifier{}".format(
            len(deprecated),
            's' * bool(deprecated)))
        for name in deprecated:
            print(name)


def run():
    import argparse
    parser = argparse.ArgumentParser(
        description='Interact with memestra cache')
    subparsers = parser.add_subparsers()

    parser.add_argument('--cache-dir', dest='cache_dir',
                        default=None,
                        action='store',
                        help='The directory where the cache is located')

    parser_set = subparsers.add_parser('set', help='Set a cache entry')
    parser_set.add_argument('--deprecated', dest='deprecated',
                            action='append',
                            help='function to flag as deprecated')
    parser_set.add_argument('--recursive', action='store_true',
                            help='set a dependency-aware cache key')
    parser_set.add_argument('input', type=str,
                            help='module.py to edit')
    parser_set.set_defaults(runner=run_set)

    parser_list = subparsers.add_parser('list', help='List cache entries')
    parser_list.set_defaults(runner=run_list)

    parser_clear = subparsers.add_parser('clear',
                                         help='Remove all cache entries')
    parser_clear.set_defaults(runner=run_clear)

    parser_docparse = subparsers.add_parser(
        'docparse',
        help='Set cache entry from docstring')
    parser_docparse.add_argument('-v', '--verbose', dest='verbose',
                                 action='store_true')
    parser_docparse.add_argument(
        '--pattern', dest='pattern', type=str, default=r'.*deprecated.*',
        help='pattern found in deprecated function docstring')
    parser_docparse.add_argument('--recursive', action='store_true',
                                 help='set a dependency-aware cache key')
    parser_docparse.add_argument('input', type=str,
                                 help='module.py to scan')
    parser_docparse.set_defaults(runner=run_docparse)

    args = parser.parse_args()
    if hasattr(args, 'runner'):
        args.runner(args)
    else:
        parser.print_help()



from memestra import memestra
from io import StringIO
from collections import namedtuple
import nbformat
import os


def nbmemestra_from_nbnode(nb, decorator, search_paths=None, reason_keyword='reason'):
    # Get code cells
    cells = nb.cells
    code_cells = [c for c in cells if c['cell_type'] == 'code']
    Cell = namedtuple('Cell', 'id begin end')
    # Aggregate code cells and generate cells list
    code_list = []
    cell_list = []
    current_line = 1
    for (num, c) in enumerate(code_cells):
        nb_lines = c['source'].count('\n') + 1
        code_list.append(c['source'])
        end_line = current_line + nb_lines
        cell_list.append(Cell(num, current_line, end_line))
        current_line = end_line
    code = '\n'.join(code_list)

    # Collect calls to deprecated functions
    deprecated_list = memestra(StringIO(code), decorator, reason_keyword,
                               search_paths)

    # Map them to cells
    result = []
    for d in deprecated_list:
        cell = next(x for x in cell_list if x.begin <= d[2] and d[2] < x.end)
        result.append((d[0],
                       'Cell[' + str(cell.id) + ']', d[2] - cell.begin + 1,
                       d[3]))

    return result


def nbmemestra(nbfile, decorator, reason_keyword='reason'):
    nb = nbformat.read(nbfile, 4)
    return nbmemestra_from_nbnode(nb, decorator,
                                  [os.path.dirname(nbfile)], reason_keyword)


def register(dispatcher):
    dispatcher['.ipynb'] = nbmemestra



import os
from setuptools import setup

kw = {"test_suite": "tests"}

dev_reqs = open("requirements-dev.txt").read().splitlines()
extras_require = {"test": dev_reqs, "dev": dev_reqs}

versionfile = os.path.join('memestra', 'version.py')
exec(open(versionfile).read())

setup(
    name="memestra",
    version=__version__,
    packages=["memestra"],
    description=__descr__,
    long_description="""
A linter that tracks reference to deprecated functions.

Memestra walks through your code base and tracks reference to function marks
with a given decorator.""",
    author="serge-sans-paille",
    author_email="serge.guelton@telecom-bretagne.eu",
    url=__url__,
    license="BSD 3-Clause",
    install_requires=open("requirements.txt").read().splitlines(),
    extras_require=extras_require,
    entry_points={'console_scripts':
                  ['memestra = memestra.memestra:run',
                   'memestra-cache = memestra.caching:run'],
                  'memestra.plugins':
                  [".ipynb = memestra.nbmemestra:register", ],
                  },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    python_requires=">=3.4",
    **kw
)
